#!/usr/bin/python

#
# KconfigSize tool--- generate database file according to BINDER and 
# INDIV result
# Copyright (C) 2005-2006 NECAS/NEC. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GPLv2, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Created by Mingwei Wang <wangmw@sh.necas.nec.com.cn>
# 2006-02-20 Modified by Mingwei Wang <wangmw@sh.necas.nec.com.cn>
# 2006-06-05 Modified by Guohuang Yin <yingh@sh.necas.nec.com.cn>
#			 Fa Wu        <wuf@sh.necas.nec.com.cn>
#            Add multiple versions support
# 2006-08-16 Modified by Guohuang Yin <yingh@sh.necas.nec.com.cn>
#            merge raw and effect db files into one single db file

import os, sys, string, re, shelve
import commands
import getopt

MAJOR_VERSION = 0
MINOR_VERSION = 2

# Default input file name generated by BINDER 
DEFAULT_BINDER_SIZERESULT = 'binder_size_result'
DEFAULT_BINDER_MEMRESULT = 'binder_mem_result'
# Default input file name generated by INDIV
DEFAULT_INDIV_SIZERESULT = 'indiv_size_result'
DEFAULT_INDIV_MEMRESULT = 'indiv_mem_result'

# Default make kernel failure flag
DEFAULT_MAKE_FAIL_f = "CP_FAIL"
# Default memory test failure flag
DEFAULT_MEMORY_FAIL_f = "MEM_FAIL"

# Default database file output directory
DEFAULT_OUTPUTDIR = "./kconfig_db"
# Default datasource input directory
DEFAULT_INPUTDIR = "."
# Default db name
db_name = ""
# Default raw item key field
DEFAULT_RAW_KEY = "-raw"
# default effect item key field
DEFAULT_EFFECT_KEY = "-effect"
# Size test data is contained in the database 
DB_has_size = 0
# Mem test data is contained in the database 
DB_has_mem = 0

# Four source filename and their file handler dictionary
configs = {}
input_fd = {}

# Directory path of source filename
# Only one-time test source files under this directory
source_dir = "" 
# shelve database file key format:
# arch-mach-KernelVersion-GccVersion-configlist-configitems-raw(effect)
# Save db key header part, as the following:
#               arch-mach-KernelVersion-GccVersion-configlist
db_key = '' 
# Default configitems and raw(effect) index in key
DEFAULT_ITEMS_KEY_FIELD_INDEX = 5
DEFAULT_EFFECT_KEY_FIELD_INDEX = 6

def usage(outfile):
	outfile.write("""Usage: %s [-c file] [-d inputdir] [-o outputdir] db_name
Output  shelve db file called db_name according
to four source files generated by INDIV and BINDER.

Meta-options:
-d, --directory    the directory where source file is contained
		   default: .
-c, --config	   config file where source file name is specified
		   default: binder_size_result, binder_mem_result
	                    indiv_size_result, indiv_mem_result
-o, --output	   output directory of database files
-h, --help         Display this help then exit.
-v, --version      Output version information then exit.
""" % sys.argv[0])

def print_error(message):
	sys.stderr.write("Error: " + message + "\n")
	sys.stderr.flush()

def error_out(message, rcode):
	print_error(message)
	sys.exit(rcode)

def read_config(info_file):
	global configs
	if info_file:
		try:
			fl = open(info_file, 'r')
		except:
			error_out("Can not open configuration file %s" %info_file,3)
		line_no = 0
		for line in fl.readlines():
			line_no += 1
			if line.startswith("#"):
				continue
			if not line.strip():
				continue
			if line.find("=") == -1:
				print_error("Syntax error in config files %s : expected '=' at line %d:\n%s" % (info_file, line_no, line))
				continue
			(name,value) = line.split('=', 1)
			name = name.strip()
			value = value.strip()
			configs[name] = value

		keys = configs.keys()
		keys.sort()
		fl.close()
	else:
		configs['BINDER_SIZE'] = DEFAULT_BINDER_SIZERESULT
		configs['BINDER_MEM'] = DEFAULT_BINDER_MEMRESULT
		configs['INDIV_SIZE'] = DEFAULT_INDIV_SIZERESULT
		configs['INDIV_MEM'] = DEFAULT_INDIV_MEMRESULT


# Search source file in the directory and save the path to source_dir
# Only search configs['BINDER_SIZE'] file currently
def search_sourcefile(dir):
	global source_dir
	global DB_has_size
	global DB_has_mem

	(status1, dirs1) = commands.getstatusoutput('find %s -name "%s" -exec dirname {} \;' % (dir, configs['BINDER_SIZE']))
	(status2, dirs2) = commands.getstatusoutput('find %s -name "%s" -exec dirname {} \;' % (dir, configs['BINDER_MEM']))

	if status1 == 0:
		dir1 = dirs1.split('\n')
		if len(dir1) == 1:
			if  dir1 != ['']: 
				DB_has_size = 1
				source_dir = dir1[0]
		else:
			error_out("Multiple %s is found in %s" %(configs['BINDER_SIZE'], dir), 1)
	if status2 == 0:
		dir2 = dirs2.split('\n')
		if len(dir2) == 1:
			if dir2 != ['']: 
				DB_has_mem = 1
				source_dir = dir2[0]
		else:
			error_out("Multiple %s is found in %s" %(configs['BINDER_MEM'], dir), 1)

	if DB_has_size == 0 and DB_has_mem == 0:
		error_out("None of %s and %s can be found in %s" \
			% (configs['BINDER_SIZE'], configs['BINDER_MEM'], dir), 1)
	if DB_has_size == 1 and DB_has_mem == 1 and dirs1 != dirs2:
		error_out("%s and %s must be in the same directory" \
			% (configs['BINDER_SIZE'], configs['BINDER_MEM']), 1)


# gen db key without config item
def gen_db_key(dir):
	global db_key
	info_items = {'ARCH': '', 'MACH': '', 'KERNEL_VERSION': '',\
				'GCC_VERSION': '', 'TARGET_LIST': ''}
	if DB_has_size == 1:
		try:
			f = open(dir + "/" + configs['BINDER_SIZE'], 'r')
		except:
			error_out("Cannot open source file %s:%s" % (dir, \
						configs['BINDER_SIZE']), 3)
	else:
		try:
			f = open(dir + "/" + configs['BINDER_MEM'], 'r')
		except:
			error_out("Cannot open source file %s:%s" % (dir, \
						configs['BINDER_MEM']), 3)

	for line in f.readlines() :
		line = line.strip()
		lines = line.split(',')
		if len(lines) != 2 :
			continue
		# get basic condition information such as ARCH, MACH, etc.
		for item in info_items.keys() :
			if lines[0] == item :
				info_items[item] = lines[1].strip()
		
	for item in info_items.keys() :
		if info_items[item] == '':
			error_out("No "+ item + " field in the " + dir + "/"\
					+ configs['BINDER_SIZE'], 2)

	db_key = info_items['ARCH'] + '-' + info_items['MACH'] + '-' +\
	info_items['KERNEL_VERSION'] + '-' + info_items['GCC_VERSION'] +\
	'-' + info_items['TARGET_LIST']
	f.close()


# open input files
def open_inputfiles(dir):
	global input_fd
	keys = configs.keys()
	keys.sort()
	for key in keys:
		f_name = dir + "/" + configs[key]
		if DB_has_size == 0 and (key == 'BINDER_SIZE' or key == 'INDIV_SIZE'):
			continue;
		if DB_has_mem == 0 and (key == 'BINDER_MEM' or key == 'INDIV_MEM'):
			continue;
		try:
			input_fd[key] = open(f_name, 'r')
		except:
			error_out("Cannot open input file:%s" % f_name, 3)


# close input files
def close_inputfiles():
	keys = input_fd.keys()
	keys.sort()
	for key in keys:
		input_fd[key].close()


# strip  '[ ]' bracket in config item
def strip_bracket(item):
	if len(item) > 2 and item[0] == '[' and item[-1] == ']':
		item=item[1:-1]
		return item
	else:
		return  item


# split a string by two with the first key which not in "()".
# parameter:
#       word:           the string you want to split.
#       key:            the key charactor to split with.
# return :
#       a list of the sub string.
def special_split(word, key):
	uneffect = 0;
	data = [];
	if word[len(word)-1] == '\n' :
		word = word[0:len(word)-1];
	start = 0;
	for i in range(0, len(word)) :
		if i == len(word)-1 :
			data.append(word[start:]);
			break;
		if word[i] == key and uneffect == 0 :
			data.append(word[start:i]);
			start = i+1;
			if start == len(word) :
				data.append("");
				break;
		elif word[i] == '[' :
			uneffect = uneffect + 1;
		elif word[i] == ']' :
			uneffect = uneffect - 1;
	return data;


# handle raw size data 	
def add_raw_size(file, db) :
	for line in file.readlines():
		line = line.strip()
		lines = special_split(line,',')
			
		if len(lines) == 5:
			config_item = lines[1].strip()
			config_item = strip_bracket(config_item)
			vmlinux = lines[3].strip()
			bzImage = lines[4].strip()
			if vmlinux == "0" and bzImage == "0":
				vmlinux = DEFAULT_MAKE_FAIL_f
				bzImage = DEFAULT_MAKE_FAIL_f
			if lines[4].strip().isalpha():
				continue
			insert_key = db_key + '-' + config_item + DEFAULT_RAW_KEY
			if DB_has_mem == 1:
				db[insert_key] = [vmlinux, bzImage]
			else:
				db[insert_key] = [vmlinux, bzImage, "", "", ""]
				
# handle raw mem data
def add_raw_mem(file, db) :
	for line in file.readlines() :
		line = line.strip()
		lines = special_split(line,',')
			
		if len(lines) == 4 or len(lines) == 6:
			if len(lines) == 4:
				config_item = lines[1].strip()
				tmpline = lines[3].strip()
				if tmpline == DEFAULT_MAKE_FAIL_f:
					used = DEFAULT_MAKE_FAIL_f
					free = DEFAULT_MAKE_FAIL_f
					total = DEFAULT_MAKE_FAIL_f
				elif tmpline == DEFAULT_MEMORY_FAIL_f:
					used = DEFAULT_MEMORY_FAIL_f
					free = DEFAULT_MEMORY_FAIL_f
					total = DEFAULT_MEMORY_FAIL_f
				else:
					continue
			else:
				config_item = lines[1].strip()
				used = lines[3].strip()
				free = lines[4].strip()
				total = lines[5].strip()
				if lines[5].strip().isalpha() :
					continue
				
			config_item = strip_bracket(config_item)
			insert_key = db_key + '-' + config_item + DEFAULT_RAW_KEY
			
			if DB_has_size == 1:
				keys = db.keys()
				if insert_key in keys :
					data = db[insert_key]
					data.append(total)
					data.append(used)
					data.append(free)
					db[insert_key] = data
			else:
				db[insert_key] = ["", "", total, used, free]


# handle effect size data
def add_effect_size(file, db) :
	for line in file.readlines() :
		line = line.strip()
		lines = special_split(line,',')
		
		if len(lines) == 3 :
			config_item = lines[0].strip()
			vmlinux = lines[1].strip()
			bzImage = lines[2].strip()
			config_item = strip_bracket(config_item)
		
			if lines[2].strip().isdigit() or lines[2].strip().find('-') == 0:
				insert_key = db_key + '-' + config_item + DEFAULT_EFFECT_KEY
				templs = config_item.split('+')
				if templs[0] == "BASE":
					temp_config_item = "NONE"
				else:
					temp_config_item = templs[0] 
				if db[db_key + '-' + temp_config_item + \
					DEFAULT_RAW_KEY][0] == DEFAULT_MAKE_FAIL_f:
					vmlinux = DEFAULT_MAKE_FAIL_f
					bzImage = DEFAULT_MAKE_FAIL_f
				if DB_has_mem == 1:
					db[insert_key] = [vmlinux, bzImage]
				else:
					db[insert_key] = [vmlinux, bzImage, "", "", ""]

def get_effect_item(keys, configitem):
	for key in keys:
		keyfield = key.split('-')
		items = keyfield[DEFAULT_ITEMS_KEY_FIELD_INDEX]
		if items == configitem:
			if keyfield[DEFAULT_EFFECT_KEY_FIELD_INDEX] == DEFAULT_EFFECT_KEY.strip('-'):
				return items
	return '0'
	
# handle effect mem data
def add_effect_mem(file, db) :
	for line in file.readlines() :
		line = line.strip()
		lines = special_split(line,',')
			
		if len(lines) == 4  :
			config_item = lines[0].strip()
			config_item = strip_bracket(config_item)
			used = lines[1].strip()
			free = lines[2].strip()
			total = lines[3].strip()
		
			templs = config_item.split('+')
			if templs[0] == "BASE":
				temp_config_item = "NONE"
			else:
				temp_config_item = templs[0] 
			m_test_f = 0
			keys = db.keys()
			if db_key + '-' + temp_config_item + DEFAULT_RAW_KEY in keys :
				if db[db_key + '-' + temp_config_item + DEFAULT_RAW_KEY \
							][2] == DEFAULT_MAKE_FAIL_f:
					used = DEFAULT_MAKE_FAIL_f
					free = DEFAULT_MAKE_FAIL_f
					total = DEFAULT_MAKE_FAIL_f
					effect_item = get_effect_item(keys, temp_config_item)
					if effect_item != '0':
						config_item = effect_item
					m_test_f = 1
				if db[db_key + '-' + temp_config_item + DEFAULT_RAW_KEY \
						][2] == DEFAULT_MEMORY_FAIL_f:
					used = DEFAULT_MEMORY_FAIL_f
					free = DEFAULT_MEMORY_FAIL_f
					total = DEFAULT_MEMORY_FAIL_f
					m_test_f = 1
			
			if lines[3].strip().isdigit() or lines[3].strip().find('-') == 0 \
				or m_test_f == 1:
				m_test_f = 0
				insert_key = db_key + '-' + config_item + DEFAULT_EFFECT_KEY
				if DB_has_size == 1:
					keys = db.keys()
					if insert_key in keys :
						data = db[insert_key]
						data.append(total)
						data.append(used)
						data.append(free)
						db[insert_key] = data
				else:
					db[insert_key] = ["", "", total, used, free]


# add one version to db
def add_version_db(dir, db):
# generate db key without config_item
	gen_db_key(dir)
# open input files
	open_inputfiles(dir)
# generate db file 
	if DB_has_size == 1:
		add_raw_size(input_fd['BINDER_SIZE'], db)
		add_effect_size(input_fd['INDIV_SIZE'], db)
	if DB_has_mem == 1:
		add_raw_mem(input_fd['BINDER_MEM'], db)
		add_effect_mem(input_fd['INDIV_MEM'], db)
# close input files
	close_inputfiles()

# gen db for all versions
def gen_db(idir, odir):

	# open db file
	try:
		db = shelve.open(odir + "/" + db_name, 'n')
	except:
		error_out("Cannot open output file %s" % db_name, 3)


	# search source file in the directory
	search_sourcefile(idir)
	add_version_db(source_dir, db)
	# close db files
	db.close()

# for debug
def show(db) :
	tmpf = open("db.log", 'w')
	list = ""
	f = shelve.open(db,'r')
	keys = f.keys()
	for key in keys :
		list = key + " "
		for i in range(len(f[key])):
			list += f[key][i] + ", "
		list += "\n"
		tmpf.write(list)
	f.close()
	tmpf.close()

def main(argv=None):
	global db_name
	inputdir = ""
	outputdir = ""
	config_name = ""

	if argv is None:
		argv = sys.argv
	try:
		opts, prog_argv = getopt.getopt(argv[1:], "o:d:c:hv",
					["help", "version",])

	except getopt.error, msg:
		sys.stderr.write("%s: %s\n" % (sys.argv[0], msg))
		sys.stderr.write("Try `%s --help' for more information\n"
				 % sys.argv[0])
		sys.exit(1)

	for opt, val in opts:
		if opt == "-h" or opt == "--help":
			usage(sys.stdout)
			sys.exit(0)

		if opt == "-v" or opt == "--version":
			sys.stdout.write("gen_db version %d.%d\n" % (MAJOR_VERSION, MINOR_VERSION))
			sys.exit(0)

		if opt == "-o" or opt == "--output":
			outputdir = val
			continue

		if opt == "-d" or opt == "--directory":
			inputdir = val
			continue

		if opt == "-c" or opt == "--config":
			config_name = val
			continue

		assert 0, "Should never get here"

# config default directory
	if len(prog_argv) == 1:
		db_name = prog_argv[0]
	else:
		sys.stderr.write("db file name?\n")
		sys.stderr.write("Try `%s --help' for more information\n"
				 % sys.argv[0])
		sys.exit(1)

	if outputdir == "":
		outputdir = DEFAULT_OUTPUTDIR
	os.system("mkdir -p " + outputdir)
	if inputdir == "":
		inputdir = DEFAULT_INPUTDIR
# read config file
	read_config(config_name)
# gen db of all versions
	gen_db(inputdir, outputdir)
# test db
#	show(outputdir + "/" + db_name)
if __name__ == "__main__":
	main()
