/*
 * KconfigSize tool---linux kernel .config generator in the bottom-up model
 * Copyright (C) 2005-2006 NEC-AS/NEC. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GPLv2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Created by Lv Zheng <zhengl@sh.necas.nec.com.cn>
 * 2006-02-20 Modified by  Fan Wu<wuf@sh.necas.nec.com.cn>
 * 2006-08-22 Modified by  Guohuang Yin<yingh@sh.nec-as.nec.com.cn>
 *             extract some funcitions shared with size2.c to size-common.c
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <getopt.h>

#define LKC_DIRECT_LINK
#include "lkc.h"
#include "size-common.h"

#define MAX_PATH 256
#define my_mkdir(x, y) mkdir(x, y)

#define DEFAULT_ARCH  "i386"
#define DEFAULT_SOURCE_TREE  "."
#define DEFAULT_OUTPUT  "."
#define DEFAULT_INDEX  "./index"
#define DEFAULT_LIST  "./list"

#define TITLE_INDEX   "index"
#define TITLE_FOCUS   "focus item"
#define TITLE_ENABLED   "enabled items"

#define SYMBOL_LOOKUP	0x10000

enum size_type {
	S_INDEPEND, S_FOCUS, S_BASIC, S_DEPEND, S_B_INDEPEND
};

#define SIZE_FIXED	0x00000001 /* fixed value, can not be changed */

union size_value {
	struct symbol *sym;
	tristate tri;
};

struct symbol_entry {
	struct symbol *sym;
	/* FIFO here */
	struct symbol_entry *next;
	struct symbol_entry *prev;
	enum size_type type; /* added for size processing */
	int flags;
	union size_value value;
	struct symbol_value user;
	struct symbol_value curr;
	int symbol_flags;
};

static struct symbol *current_symbol = NULL;

struct symbol_entry current_list;
struct symbol_entry config_list;
struct symbol_entry temp_list;


const char *progname = NULL;
char *src_tree = NULL;
const char *menu_output = NULL;
char *config_set = NULL;
char * index_file = NULL;
char * list_file = NULL;
static int valid_stdin = 1;

static tristate modules_val;

static char nohelp_text[] = N_("Sorry, no help available for this option yet.\n");

static struct option long_options[] = {
	{"help", 0, 0, 'h'},
	{"arch", 1, 0, 'a'},
	{"kernel", 1, 0, 'k'},
	{"dir", 1, 0, 'd'},
	{"index", 1, 0, 'i'},
	{"config", 1, 0, 'c'},
	{0, 0, 0, 0}
};

#define sym_is_depend(sym) sym_has_size_type(&config_list, sym, S_DEPEND)
#define sym_is_focus(sym) sym_has_size_type(&config_list, sym, S_FOCUS)
#define sym_is_basic(sym) sym_has_size_type(&config_list, sym, S_BASIC)
#define sym_is_sizing(sym) sym_has_size_type(&current_list, sym, S_INDEPEND)

#define for_all_selects(sym, st) for_all_properties(sym, st, P_SELECT)

#define for_all_forward_entries(pos, n, head)			\
	for (pos = (head)->next, n = pos->next;			\
		pos != (head); pos = n, n = pos->next)
#define for_all_backward_entries(pos, n, head)			\
	for (pos = (head)->prev, n = pos->prev;			\
		pos != (head); pos = n, n = pos->prev)
#define for_all_list_entries(pos, head)				\
	for (pos = (head)->next; pos != (head); pos = pos->next)
#define for_all_size_symbols(list)				\
	for_all_list_entries(list, &current_list)
#define for_all_sizing_symbols(list)				\
	for_all_list_entries(list, &current_list)		\
		if (sym_is_sizing(list->sym))
#define for_all_depend_symbols(list)				\
	for_all_list_entries(list, &config_list)		\
		if (sym_is_depend(list->sym))
#define for_all_focus_symbols(list)				\
	for_all_list_entries(list, &config_list)		\
		if (sym_is_focus(list->sym))
#define for_all_basic_symbols(list)				\
	for_all_list_entries(list, &config_list)		\
		if (sym_is_basic(list->sym))

int size_conf_choice(struct menu *menu);

static bool sym_has_size_type(struct symbol_entry *list, struct symbol *sym, enum size_type type);
static bool size_make_sym(struct symbol *sym, tristate val);
static bool size_make_equal(struct symbol *sym1, struct symbol *sym2, enum expr_type type);
static bool size_make_expr_yes(struct expr *expr, bool fast);
static bool size_make_expr_no(struct expr *expr, bool fast);
static void size_reset_all_sym();
static void size_usage();

struct symbol_entry *size_find_sym(struct symbol_entry *list, struct symbol *sym)
{
	struct symbol_entry *sym_list, *tmp;

	if (!sym)
		return NULL;

	for_all_forward_entries(sym_list, tmp, list) {
		if (sym_list->sym == sym)
			return sym_list;
	}
	return NULL;
}

void size_init_sym_lists()
{
	current_list.next = current_list.prev = &current_list;
	config_list.next = config_list.prev = &config_list;
}

struct symbol_entry *size_lookup_sym(struct symbol_entry *head, struct symbol *sym)
{
	struct symbol_entry *entry;

	entry = size_find_sym(head, sym);
	if (!entry) {
		entry = malloc(sizeof (*entry));
		memset(entry, 0, sizeof (*entry));
		entry->sym = sym;
		entry->type = S_INDEPEND;
		head->prev->next = entry;
		entry->prev = head->prev;
		head->prev = entry;
		entry->next = head;
	}
	return entry;
}

static void size_remove_sym(struct symbol_entry *entry)
{
	if (!entry)
		return;

	entry->next->prev = entry->prev;
	entry->prev->next = entry->next;
	free(entry);
}

void size_reset_sym_list()
{
	struct symbol_entry *sym_list, *tmp;

	for_all_forward_entries(sym_list, tmp, &current_list) {
		size_remove_sym(sym_list);
	}
}

static bool sym_has_size_type(struct symbol_entry *list, struct symbol *sym,
			      enum size_type type)
{
	struct symbol_entry *sym_list;

	sym_list = size_find_sym(list, sym);
	if (sym_list && sym_list->type == type)
		return TRUE;
	return FALSE;
}

static void size_make_basic()
{
	struct symbol_entry *sym_list;
	struct symbol_entry *sym_warn;

	for_all_basic_symbols(sym_list) {
		if (sym_list && sym_list->sym)
			size_make_sym(sym_list->sym, yes);
	}

	/* find implied basic symbols */
	for_all_sizing_symbols(sym_list) {
		sym_list->type = S_BASIC;
	}
}

static void size_reset_config()
{
	size_reset_all_sym();
	size_make_basic();
	do {
		conf_cnt = 0;
		size_conf_check(&rootmenu);
	} while (conf_cnt);
}

static bool size_read_config(const char *name)
{
	FILE *in = NULL;
	char line[1024];
	char *p, *p2;
	int lineno = 0;
	struct symbol *sym;
	struct property *prop;
	struct expr *e;
	int i;
	struct symbol_entry *sym_list;

	if (name) {
		in = zconf_fopen(name);
	} else {
		return FALSE;
	}

	if (!in)
		return FALSE;

	while (fgets(line, sizeof(line), in)) {
		lineno++;
		sym = NULL;
		switch (line[0]) {
		case '#':
			if (memcmp(line + 2, "CONFIG_", 7))
				continue;
			p = strchr(line + 9, ' ');
			if (!p)
				continue;
			*p++ = 0;
			if (strncmp(p, "is not set", 10))
				continue;
			sym = sym_find(line + 9);
			if (!sym) {
				fprintf(stderr, "%s:%d: trying to assign nonexistent symbol %s\n", name, lineno, line + 9);
				break;
			}
			switch (sym->type) {
			case S_BOOLEAN:
			case S_TRISTATE:
				sym->user.tri = no;
				sym->flags &= ~SYMBOL_NEW;
				break;
			default:
				;
			}
			break;
		case 'C':
			if (memcmp(line, "CONFIG_", 7))
				continue;
			p = strchr(line + 7, '=');
			if (!p)
				continue;
			*p++ = 0;
			p2 = strchr(p, '\n');
			if (p2)
				*p2 = 0;
			sym = sym_find(line + 7);
			if (!sym) {
				fprintf(stderr, "%s:%d: trying to assign nonexistent symbol %s\n", name, lineno, line + 7);
				break;
			}
			/* load default string/int/hex value */
			switch (sym->type) {
			case S_BOOLEAN:
			case S_TRISTATE:
				if (p[0] == 'b') {
					sym_list = size_lookup_sym(&config_list, sym);
					if (sym_list)
						sym_list->type = S_BASIC;
				} else if (p[0] == 'f') {
					sym_list = size_lookup_sym(&config_list, sym);
					if (sym_list)
						sym_list->type = S_FOCUS;
				} else if (p[0] == 'd') {
					sym_list = size_lookup_sym(&config_list, sym);
					if (sym_list)
						sym_list->type = S_DEPEND;
				}
				break;
			case S_STRING:
				if (*p++ != '"')
					break;
				for (p2 = p; (p2 = strpbrk(p2, "\"\\")); p2++) {
					if (*p2 == '"') {
						*p2 = 0;
						break;
					}
					memmove(p2, p2 + 1, strlen(p2));
				}
				if (!p2) {
					fprintf(stderr, "%s:%d: invalid string found\n", name, lineno);
					exit(1);
				}
			case S_INT:
			case S_HEX:
				if (sym_string_valid(sym, p)) {
					sym->user.val = strdup(p);
					sym->curr.val = sym->user.val;
					sym->flags &= ~SYMBOL_NEW;
				} else {
					fprintf(stderr, "%s:%d: symbol value '%s' invalid for %s\n", name, lineno, p, sym->name);
					exit(1);
				}
				break;
			default:
				;
			}
			break;
		case '\n':
			break;
		default:
			continue;
		}
	}
	fclose(in);

	return TRUE;
}

void size_reset_value(struct symbol *sym)
{
	char *newval;
	struct symbol_entry *sym_list;

	if (sym_is_const(sym))
		return;

	sym_list = size_find_sym(&current_list, sym);

	if (sym_list) {
		sym->curr = sym_list->curr;
		sym->user = sym_list->user;
		sym->flags = sym_list->symbol_flags;
	} else {
		sym->flags |= SYMBOL_NEW | SYMBOL_CHANGED;
		sym->flags &= ~SYMBOL_VALID;
		switch (sym->type) {
		case S_INT:
		case S_HEX:
		case S_STRING:
			if (sym->user.val) {
				sym->curr.val = sym->user.val;
				sym->flags &= ~SYMBOL_NEW;
			}
			break;
		default:
			sym->user.val = NULL;
			sym->user.tri = no;
		}
	}
}

int size_conf_choice(struct menu *menu)
{
	struct symbol *sym, *def_sym;
	struct menu *child;
	int type;
	bool is_new;

	sym = menu->sym;
	type = sym_get_type(sym);
	is_new = !sym_has_value(sym);
	if (sym_is_changable(sym)) {
		size_conf_sym(menu);
		sym_calc_value(sym);
		switch (size_get_tristate_value(sym)) {
		case no:
			return TRUE;
		case mod:
			return FALSE;
		case yes:
			break;
		}
	} else {
		switch (size_get_tristate_value(sym)) {
		case no:
			return TRUE;
		case mod:
			return FALSE;
		case yes:
			break;
		}
	}

	while (1) {
		int cnt, def;
		def_sym = sym_get_choice_value(sym);
		cnt = def = 0;
		line[0] = '0';
		line[1] = 0;
		for (child = menu->list; child; child = child->next) {
			if (!menu_is_visible(child))
				continue;
			if (!child->sym) {
				continue;
			}
			cnt++;
			if (child->sym == def_sym) {
				def = cnt;
			}
		}
		if (cnt == 1) {
			goto conf_childs;
		}
		if (!is_new) {
			cnt = def;
			sprintf(line, "%d\n", cnt);
			goto conf_childs;
		} else {
			cnt = def;
			sprintf(line, "%d\n", cnt);
			goto conf_childs;
		}
		strip(line);
		if (!line[0])
			cnt = def;
		else if (isdigit(line[0]))
			cnt = atoi(line);
		else
			continue;

conf_childs:
		for (child = menu->list; child; child = child->next) {
			if (!child->sym || !menu_is_visible(child))
				continue;
			if (!--cnt)
				break;
		}
		if (!child)
			continue;

		size_set_choice_value(sym, child->sym);
		printf("%s\n",child->sym->name);
		if (child->list) {
			indent += 2;
			size_conf_menu(child->list);
			indent -= 2;
		}
		return TRUE;
	}
}

static bool size_make_equal(struct symbol *sym1, struct symbol *sym2, enum expr_type type)
{
	tristate val1, val2;

	assert(type == E_EQUAL || type == E_UNEQUAL);

	val1 = size_get_tristate_value(sym1);
	val2 = size_get_tristate_value(sym2);

	if ((type == E_EQUAL && val1 == val2) || (type == E_UNEQUAL && val1 != val2))
		return TRUE;
	else if (type == E_EQUAL) {
		if (sym_is_const(sym1))
			return size_make_sym(sym2, val1);
		else if (sym_is_const(sym2))
			return size_make_sym(sym1, val2);
		else if (val1 > val2)
			return size_make_sym(sym2, val1);
		else
			return size_make_sym(sym1, val2);
	} else {
		if (sym_is_const(sym1))
			return size_make_sym(sym2, reverse_value(val1));
		else if (sym_is_const(sym2))
			return size_make_sym(sym1, reverse_value(val2));
		else /* val1 == val2 */
			return size_make_sym(sym1, reverse_value(val1));
	}
}

static bool size_make_expr_yes(struct expr *expr, bool fast)
{
	const char *str1, *str2;
	int val1, val2;
	struct symbol_entry *sym_list;

	if (!expr)
		return TRUE;

	if (size_calc_expr(expr) == yes)
		return TRUE;

	switch (expr->type) {
	case E_SYMBOL:
		if (fast) {
			for_all_size_symbols(sym_list) {
				if (sym_list && sym_list->sym->name &&
					expr->left.sym && expr->left.sym->name &&
					!strcmp(sym_list->sym->name, expr->left.sym->name)) {
					return TRUE;
				}
			}
			return FALSE;
		}
		return size_make_sym(expr->left.sym, yes);
	case E_AND:
		if (!size_make_expr_yes(expr->left.expr, fast)) {
			return FALSE;
		}
		if (!size_make_expr_yes(expr->right.expr, fast)) {
			return FALSE;
		}
		return TRUE;
	case E_OR:
		if (!size_make_expr_yes(expr->left.expr, fast)) {
			return size_make_expr_yes(expr->right.expr, fast);
		}
		return TRUE;
	case E_NOT:
		return size_make_expr_no(expr->left.expr, fast);
	case E_EQUAL:
	case E_UNEQUAL:
		return size_make_equal(expr->left.sym, expr->right.sym, expr->type);
	default:
		printf("size_make_expr_yes: %d?\n", expr->type);
	}
	return TRUE;
}

static bool size_make_expr_no(struct expr *expr, bool fast)
{
	const char *str1, *str2;
	int val1, val2;
	struct symbol_entry *sym_list;

	if (!expr)
		return TRUE;

	if (size_calc_expr(expr) == no)
		return TRUE;

	switch (expr->type) {
	case E_SYMBOL:
		if (fast) {
			for_all_size_symbols(sym_list) {
				if (sym_list && sym_list->sym->name &&
					expr->left.sym && expr->left.sym->name &&
					!strcmp(sym_list->sym->name, expr->left.sym->name)) {
					return FALSE;
				}
			}
		}
		return size_make_sym(expr->left.sym, no);
	case E_AND:
		if (!size_make_expr_no(expr->left.expr, fast)) {
			return size_make_expr_no(expr->right.expr, fast);
		}
		return TRUE;
	case E_OR:
		if (!size_make_expr_no(expr->left.expr, fast)) {
			return FALSE;
		}
		if (!size_make_expr_no(expr->right.expr, fast)) {
			return FALSE;
		}
		return TRUE;
	case E_NOT:
		return size_make_expr_yes(expr->left.expr, fast);
	case E_EQUAL:
	case E_UNEQUAL:
		return size_make_equal(expr->left.sym, expr->right.sym, reverse_expr(expr->type));
	default:
		printf("size_make_expr_no: %d?\n", expr->type);
	}
	return TRUE;
}

static bool size_make_sym(struct symbol *sym, tristate val)
{
	struct menu *menu = NULL, *child = NULL;
	struct property *prop;
	struct symbol_entry *sym_list;
	struct symbol *choice_val = NULL;
	struct symbol *sym_tmp= NULL;

	if (!sym)
		return TRUE;

	/* we got choice entry's value, remember it into the list */
	if (sym_is_choice_value(sym)) {
		//printf("choice: %s\n", sym->name);
		choice_val = sym;
		sym = prop_get_symbol(sym_get_choice_prop(sym));
	}

	/* get menu entry */
	for_all_prompts(sym, prop) {
		menu = prop->menu;
		if (menu) break;
	}

	sym_list = size_find_sym(&current_list, sym);
	if (!sym_list) {
		struct symbol_entry *tmp;

		/* we don't care choice entry unless we got it's value */
		if (sym_is_choice(sym) && !choice_val)
			return TRUE;

		sym_list = size_lookup_sym(&current_list, sym);
		sym_list->flags = 0;

		tmp = size_find_sym(&config_list, sym);

		if (choice_val) {
			sym_list->value.sym = choice_val;
		} else {
			sym_list->value.tri = val;
		}

		sym_list->user = sym->user;
		sym_list->curr = sym->curr;
		sym_list->symbol_flags = sym->flags;
	} else {
		if ((choice_val && sym_list->value.sym != choice_val) ||
		    (!choice_val && sym_list->value.tri != val)) {
			/* collision detected */

			if (choice_val)
				printf("warning: collision of <choice> %s <-> %s\n",
					choice_val->name, sym_list->value.sym->name);
			else if (sym_list->sym && sym_list->sym->name)
				printf("warning: collision of %s\n", sym_list->sym->name);
			else if (sym_is_choice(sym)) {
				printf("warning: collision of <choice>\n");
			} else
				printf("warning: collision of <unknown>\n");

			return FALSE;
		}
	}

	if (menu) {
		if (!choice_val) {
			if (!size_make_expr_yes(menu->dep, TRUE)) {
				if (!size_make_expr_yes(menu->dep, FALSE)) {
					return FALSE;
				}
			}
		}
	} else {
			
		
		if (!size_make_expr_yes(sym->rev_dep.expr, TRUE)) {
			if (!size_make_expr_yes(sym->rev_dep.expr, FALSE)) {
				return FALSE;
			}
		}
	}

	if (choice_val) {
		size_set_choice_value(sym, choice_val);
	} else {
		switch (sym_get_type(sym)) {
		case S_INT:
		case S_HEX:
		case S_STRING:
			break;
		case S_TRISTATE:
		case S_BOOLEAN:
			size_set_tristate_value(sym, val);
			break;
		default:
			;
		}
	}
	/* recalc visibility */
	sym_calc_value(sym);

	/* output all selects configurations */
	if (val == yes) {
		for_all_selects(sym, prop) {
			prop->visible.tri = expr_calc_value(prop->visible.expr);
			//printf("sym is %s, visible is %d\n",prop_get_symbol(prop)->name,prop->visible.tri);
			if (prop->visible.tri == yes) {
				sym_tmp = prop_get_symbol(prop);
				if (!size_make_sym(sym_tmp, yes))
					return FALSE;
			}
		}
	}

	return TRUE;
}

static void size_reset_all_sym()
{
	int i;
	struct symbol *sym;
	struct symbol_entry *sym_list;

	for_all_symbols(i, sym) {
		size_reset_value(sym);
	}
	size_reset_sym_list();
}

static void size_output_depends(FILE *fp_index, struct symbol_entry *focus_sym)
{
	struct symbol_entry *sym_list = NULL;
	for_all_sizing_symbols(sym_list) {
		if (sym_list->sym->name){
			if (strcmp(sym_list->sym->name, focus_sym->sym->name)) {
				fprintf(fp_index, "+%s", sym_list->sym->name);
			}
		} else {
			/* choice value */
			if (sym_list->value.sym && strcmp(sym_list->value.sym->name, focus_sym->sym->name)) {
				fprintf(fp_index, "+%s", sym_list->value.sym->name);
				//printf("hi: %s\n",sym_list->value.sym->name);
			}
		}
	}

	fprintf(fp_index,"\n");
}

/*
 * add the default symbol
 */
static void size_add_default_sym()
{
	struct symbol *sym;
	struct menu *menu;
	int type;
	int i;
	struct symbol_entry *sym_list;

	//SYMBOL_LOOKUP is added just for this function to instead of SYMBOL_WRITE
	for_all_symbols(i,sym) {
		sym->flags |= SYMBOL_LOOKUP;
	}

	sym = sym_lookup("KERNELRELEASE", 0);
	sym_calc_value(sym);
	sym_clear_all_valid();

	menu = rootmenu.list;
	while (menu) {
		sym = menu->sym;
		if (!sym) {
			if (!menu_is_visible(menu))
				goto next;
			menu_get_prompt(menu);
		} else if (!(sym->flags & SYMBOL_CHOICE)) {
			sym_calc_value(sym);
			if (!(sym->flags & SYMBOL_LOOKUP))
				goto next;
			sym->flags &= ~SYMBOL_LOOKUP;
			type = sym->type;
			if (type == S_TRISTATE) {
				sym_calc_value(modules_sym);
				if (modules_sym->curr.tri == no)
					type = S_BOOLEAN;
			}
			switch (type) {
			case S_BOOLEAN:
			case S_TRISTATE:
				switch (sym_get_tristate_value(sym)) {
				case no:
					break;
				case mod:
					break;
				case yes:
					sym_list = size_lookup_sym(&current_list, sym);
					sym_list->value.tri = yes;
					break;
				}
				break;
			case S_STRING:
				sym_get_string_value(sym);
				break;
			case S_HEX:
				sym_get_string_value(sym);
				break;
			case S_INT:
				sym_get_string_value(sym);
				break;
			}
		}

	next:
		if (menu->list) {
			menu = menu->list;
			continue;
		}
		if (menu->next)
			menu = menu->next;
		else while ((menu = menu->parent)) {
			if (menu->next) {
				menu = menu->next;
				break;
			}
		}
	}
}


static void size_make_focus_sym(FILE *fp_index, FILE *fp_list, struct symbol_entry *focus_sym)
{
	//add the default symbols enabled by size_make_basic
	size_add_default_sym();
	struct symbol_entry *list;
	struct symbol_entry *list_current;
	int need_delete;
	struct symbol_entry *sym_list, *tmp;

	list = current_list.next;
	if(list == NULL)
		fprintf(stderr, "list is null");
	temp_list.next = temp_list.prev = &temp_list;
	for(list;list!=&current_list;list = list->next){
		size_lookup_sym(&temp_list,list->sym);
	}

/*
	struct symbol_entry *sym_temp;
	for_all_sizing_symbols(sym_temp) {
		sym_temp->type = S_B_INDEPEND;
	}
*/
	if (!size_make_sym(sym_find(focus_sym->sym->name), yes)) {
		fprintf(stderr, _("\n*** Error(%d) during calculating focus symbol %s.\n\n"),focus_no, focus_sym->sym->name);
	}
	do {
		conf_cnt = 0;
		size_conf_check(&rootmenu);
	} while (conf_cnt);

	//add the default symbols enabled by size_make_sym
	size_add_default_sym();

	list = temp_list.next;
	for(list;list!=&temp_list;list = list->next){
		need_delete =0;
		list_current = current_list.next;
		for(list_current;list_current!=&current_list;list_current = list_current->next){
			if(list_current->sym == list->sym){
				need_delete =1;
				break;
			}
		}
		if(need_delete)
			size_remove_sym(list_current);
	}

	for_all_forward_entries(sym_list, tmp, &temp_list) 
		size_remove_sym(sym_list);
}

/*
 * Generate output files
 */
static void size_make_focus()
{
	FILE *fp_index, *fp_list;
	struct symbol_entry *focus_list = NULL;
	struct symbol_entry *output_list = NULL;
	char tmp_path[MAX_PATH];

	focus_no = 0;
	fp_index = fopen(index_file, "w");
	if (!fp_index) {
		perror("fopen: index file");
		exit(1);
	}

	/* list file is used to hold .config file path info */
	fp_list = fopen(list_file,"w");
	if (!fp_list) {
		perror("fopen: list file");
		exit(1);
	}

	size_reset_config();
	/* add title to the index file */
	fprintf(fp_index, "%s,%s,%s\n", TITLE_INDEX,TITLE_FOCUS,TITLE_ENABLED);
	fprintf(fp_index, "%d,NONE,NONE\n", focus_no);

	size_gen_config();
	/* write a line into list file, absolute path is needed */
	fprintf(fp_list,"%d,%s" DIRDELIM "%d.config\n",focus_no,absolute_result_path,focus_no);

	for_all_focus_symbols(focus_list) {
		focus_no++;
		/* reset config file */
		size_reset_config();
		
		size_make_focus_sym(fp_index, fp_list, focus_list);

		fprintf(fp_index, "%d,%s,%s", focus_no, focus_list->sym->name, focus_list->sym->name);
		size_output_depends(fp_index, focus_list);

		size_gen_config();
		/* write a line into list file, absolute path is needed */
		fprintf(fp_list,"%d,%s" DIRDELIM "%d.config\n",focus_no,absolute_result_path,focus_no);
	}
	size_reset_sym_list();

	fclose(fp_index);
	fclose(fp_list);
}

int main(int argc, char **argv)
{
	int c = 0;
	const char *kconfig_file;
	FILE *out_file = NULL, *out;
	char *arch = NULL;
	char arch_kconfig[256];

	progname = basename(argv[0]);
	size_init_sym_lists();

	while (1) {
		int option_index = 0;
		c = getopt_long(argc, argv, "a:k:d:i:c:m:h", long_options, &option_index);
		if (c == -1)
			break;
		
		switch (c) {
		case 'a':
			arch = strdup(optarg);
			break;
		case 'k':
			src_tree = strdup(optarg);
			setenv(SRCTREE, (char *)src_tree, 1);
			break;
		case 'd':
			result_path = strdup(optarg);
			break;
		case 'i':
			index_file = strdup(optarg);
			break;
		case 'c':
			list_file = strdup(optarg);
			break;
		case 'm':
			menu_output = optarg;
			break;
		case 'h':
			size_usage();
			exit(0);
		case '?':
		default:
			size_usage();
			exit(1);
		}
	}

	if (optind < argc) {
		config_set = strdup(argv[optind]);
	} else {
		fprintf(stderr, "Missing config list file....\n");
		size_usage();
		exit(1);
	}

	/* if not set, use default value */
	if(arch == NULL)
		arch=DEFAULT_ARCH;
	if(src_tree == NULL)
		src_tree=DEFAULT_SOURCE_TREE;
	if(result_path == NULL)
		result_path=DEFAULT_OUTPUT;
	if(index_file == NULL)
		index_file=combine_path(result_path,DEFAULT_INDEX);
	if(list_file == NULL)
		list_file=combine_path(result_path,DEFAULT_LIST);

	/* get Kconfig path */
	memset(arch_kconfig,0,sizeof(arch_kconfig));
	strcpy(arch_kconfig,"arch");
	strcat(arch_kconfig, DIRDELIM);
	strcat(arch_kconfig,arch);
	strcat(arch_kconfig, DIRDELIM);
	strcat(arch_kconfig,"Kconfig");
	kconfig_file = combine_path(src_tree, arch_kconfig);

	/* if output directory doesn't exist, create it */
	if(access(result_path, F_OK) != 0){
		if(mkdir(result_path,0755)!= 0){
			perror("create output directory failure");
			exit(1);
		}	
	}

	get_current_path();
	if(result_path[0] == DIRDELIMC){
		strcpy(absolute_result_path,result_path);
	} else {
		strcat(absolute_result_path, DIRDELIM);
		strcat(absolute_result_path, result_path);
	}
	valid_stdin = isatty(0) && isatty(1) && isatty(2);
	conf_parse(kconfig_file);

	if (menu_output) {
		out = stdout;
		if (strcmp(menu_output, "-")) {
			if ((out_file = fopen(menu_output, "w")) == NULL) {
				perror("fopen");
				exit(1);
			}
			out = out_file;
		}
		size_print_menu(out);
		if (out_file)
			fclose(out_file);
	}

	size_read_config(config_set);
	size_make_focus();
	return 0;
}

static void size_usage()
{
	fprintf(stderr, "Usage %s <item_file> [options] \n", progname?progname:"GENCONF");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "        -a, specify arch\n");
	fprintf(stderr, "        -k, specify kernel source tree\n");
	fprintf(stderr, "        -d, specify results output directory\n");
	fprintf(stderr, "        -i, specify index file\n");
	fprintf(stderr, "        -c, specify config list file\n");
	fprintf(stderr, "        -m, specify category file\n");
}

