/*
 * KconfigSize tool---linux kernel .config generator in the top-down model
 * Copyright (C) 2005-2006 NEC-AS/NEC. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GPLv2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Created by Lv Zheng <zhengl@sh.necas.nec.com.cn>
 * 2006-02-20 Modified by  Fan Wu<wuf@sh.necas.nec.com.cn>
 * 2006-07-10 Modified by  Fan Wu<wuf@sh.necas.nec.com.cn>
 *            1. Disable one config item instead of enabling one config item
 *            2. input interface changed to .config file
 * 2006-08-22 Modified by  Guohuang Yin<yingh@sh.nec-as.nec.com.cn>
 *             extract some funcitions shared with size1.c to size-common.c
 */			   

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <getopt.h>

#define LKC_DIRECT_LINK
#include "lkc.h"
#include "size-common.h"

#define MAX_PATH 256

#define DEFAULT_ARCH  "i386"
#define DEFAULT_SOURCE_TREE  "."
#define DEFAULT_OUTPUT  "."
#define DEFAULT_INDEX  "./index"
#define DEFAULT_LIST  "./list"

#define TITLE_INDEX   "index"
#define TITLE_FOCUS   "focus item"
#define TITLE_ENABLED   "enabled items"

#define SYMBOL_LOOKUP	0x10000

enum size_type {
	S_INDEPEND, S_FOCUS, S_BASIC, S_DEPEND, S_B_INDEPEND, S_MENU
};

union size_value {
	struct symbol *sym;
	tristate tri;
};

struct symbol_entry {
	struct symbol *sym;
	/* used only when type is S_MENU */
	char menu_name[256];
	/* FIFO here */
	struct symbol_entry *next;
	struct symbol_entry *prev;
	enum size_type type; /* added for size processing */
	int flags;
	union size_value value;
	struct symbol_value user;
	struct symbol_value curr;
	int symbol_flags;
};

static struct symbol *current_symbol = NULL;
struct symbol_entry config_list;
struct symbol_entry focus_list;

const char *progname = NULL;
char *src_tree = NULL;
const char *menu_output = NULL;
char *config_set = NULL;
char *focus_set = NULL;
char * index_file = NULL;
char * list_file = NULL;
static int valid_stdin = 1;

static tristate modules_val;

static char nohelp_text[] = N_("Sorry, no help available for this option yet.\n");

static struct option long_options[] = {
	{"help", 0, 0, 'h'},
	{"arch", 1, 0, 'a'},
	{"kernel", 1, 0, 'k'},
	{"dir", 1, 0, 'd'},
	{"focus", 1, 0, 'f'},
	{"index", 1, 0, 'i'},
	{"config", 1, 0, 'c'},
	{0, 0, 0, 0}
};

#define for_all_selects(sym, st) for_all_properties(sym, st, P_SELECT)

#define for_all_forward_entries(pos, n, head)			\
	for (pos = (head)->next, n = pos->next;			\
		pos != (head); pos = n, n = pos->next)
#define for_all_backward_entries(pos, n, head)			\
	for (pos = (head)->prev, n = pos->prev;			\
		pos != (head); pos = n, n = pos->prev)
#define for_all_list_entries(pos, head)				\
	for (pos = (head)->next; pos != (head); pos = pos->next)
#define for_all_focus_symbols(list)				\
	for_all_list_entries(list, &focus_list)		
#define for_all_config_symbols(list)				\
	for_all_list_entries(list, &config_list)		

int size_conf_choice(struct menu *menu);

static bool sym_has_size_type(struct symbol_entry *list, struct symbol *sym, enum size_type type);

static bool size_make_sym(struct symbol *sym, tristate val);
static void size_reset_all_sym();
static void size_usage();
static void size_remove_sym(struct symbol_entry *entry);
static int copy_asc_file(const char *from, const char *to);
static struct symbol_entry *get_sub_focus_list(struct symbol_entry * focus);
static bool focus_is_menu(const char *line, char **menu_item);
static struct symbol_entry *size_add_menu(struct symbol_entry *head, char *menu);


void size_init_lists()
{
	config_list.next = config_list.prev = &config_list;
	focus_list.next = focus_list.prev = &focus_list;
}

void size_cleanup()
{
	struct symbol_entry *sym_list, *tmp;

	for_all_forward_entries(sym_list, tmp, &config_list) {
		size_remove_sym(sym_list);
	}
	
	for_all_forward_entries(sym_list, tmp, &focus_list) {
		size_remove_sym(sym_list);
	}
}

static struct symbol_entry *size_add_menu(struct symbol_entry *head, char *menu)
{
	struct symbol_entry *entry;

	if(menu == NULL)
		return;
	
	entry = malloc(sizeof (*entry));
	memset(entry, 0, sizeof (*entry));
	if (strlen(menu) < 254)
		strcpy(entry->menu_name,menu);
	entry->type = S_MENU;
	head->prev->next = entry;
	entry->prev = head->prev;
	head->prev = entry;
	entry->next = head;
	return entry;
}

struct symbol_entry *size_add_sym(struct symbol_entry *head, struct symbol *sym)
{
	struct symbol_entry *entry;

	if(sym == NULL)
		return;
	
	entry = malloc(sizeof (*entry));
	memset(entry, 0, sizeof (*entry));
	entry->sym = sym;
	head->prev->next = entry;
	entry->prev = head->prev;
	head->prev = entry;
	entry->next = head;
	return entry;
}

static void size_remove_sym(struct symbol_entry *entry)
{
	if (!entry)
		return;

	entry->next->prev = entry->prev;
	entry->prev->next = entry->next;
	free(entry);
}

static bool focus_is_menu(const char *line, char **menu_item)
{
	char *p = NULL;
	int i = 0;
	if (!line)
		return FALSE;
	if (memcmp(line, "MENU", 4))
		return FALSE;
	p = (char *)(line + 4);
	while (*(p + i) ==' ') {
		++i;
	}
	*menu_item = p + i;

	p = strchr(*menu_item, '\n');
	if (p)
		*p = 0;
	strip(*menu_item);
	return TRUE;
}

/* read focus items from focus file and add it to focus_list */
static bool size_read_focus(const char *name)
{
	FILE *in = NULL;
	char line[1024];
	char *p;
	char *focus_menu = NULL;
	int lineno = 0;
	struct symbol *sym;

	if (name) {
		in = zconf_fopen(name);
	} else {
		return FALSE;
	}

	if (!in)
		return FALSE;

	while (fgets(line, sizeof(line), in)) {
		lineno++;
		sym = NULL;
		switch (line[0]) {
		case 'C':
			if (memcmp(line, "CONFIG_", 7))
				continue;
			p = strchr(line + 7, '\n');
			if (p)
				*p = 0;
			sym = sym_find(line + 7);
			if (!sym) {
				fprintf(stderr, "%s:%d: trying to assign nonexistent symbol %s\n", name, lineno, line + 7);
				break;
			}
			
			/* load default string/int/hex value */
			switch (sym->type) {
			case S_BOOLEAN:
			case S_TRISTATE:
				size_add_sym(&focus_list, sym);
				break;
			default:
				break;
			}
			break;
		case '#':
			break;
			
		default:
			if (focus_is_menu(line, &focus_menu))
				size_add_menu(&focus_list, focus_menu);
			break;
		}
	}
	fclose(in);

	return TRUE;
}

static bool size_read_config(const char *name)
{
	FILE *in = NULL;
	char line[1024];
	char *p, *p2;
	int lineno = 0;
	struct symbol *sym;
	struct property *prop;
	struct expr *e;
	int i;
	struct symbol_entry *sym_list;

	if (name) {
		in = zconf_fopen(name);
	} else {
		return FALSE;
	}

	if (!in)
		return FALSE;

	while (fgets(line, sizeof(line), in)) {
		lineno++;
		sym = NULL;
		switch (line[0]) {
		case '#':
			if (memcmp(line + 2, "CONFIG_", 7))
				continue;
			p = strchr(line + 9, ' ');
			if (!p)
				continue;
			*p++ = 0;
			if (strncmp(p, "is not set", 10))
				continue;
			sym = sym_find(line + 9);
			if (!sym) {
				fprintf(stderr, "%s:%d: trying to assign nonexistent symbol %s\n", name, lineno, line + 9);
				break;
			}
			switch (sym->type) {
			case S_BOOLEAN:
			case S_TRISTATE:
				sym->user.tri = no;
				sym->flags &= ~SYMBOL_NEW;
				break;
			default:
				;
			}
			break;
		case 'C':
			if (memcmp(line, "CONFIG_", 7))
				continue;
			p = strchr(line + 7, '=');
			if (!p)
				continue;
			*p++ = 0;
			p2 = strchr(p, '\n');
			if (p2)
				*p2 = 0;
			sym = sym_find(line + 7);
			if (!sym) {
				fprintf(stderr, "%s:%d: trying to assign nonexistent symbol %s\n", name, lineno, line + 7);
				break;
			}
			
			/* load default string/int/hex value */
			switch (sym->type) {
			case S_BOOLEAN:
			case S_TRISTATE:
				size_add_sym(&config_list, sym);
				break;
			case S_STRING:
				if (*p++ != '"')
					break;
				for (p2 = p; (p2 = strpbrk(p2, "\"\\")); p2++) {
					if (*p2 == '"') {
						*p2 = 0;
						break;
					}
					memmove(p2, p2 + 1, strlen(p2));
				}
				if (!p2) {
					fprintf(stderr, "%s:%d: invalid string found\n", name, lineno);
					exit(1);
				}
			case S_INT:
			case S_HEX:
				if (sym_string_valid(sym, p)) {
					sym->user.val = strdup(p);
					sym->curr.val = sym->user.val;
					sym->flags &= ~SYMBOL_NEW;
				} else {
					fprintf(stderr, "%s:%d: symbol value '%s' invalid for %s\n", \
							name, lineno, p, sym->name);
					exit(1);
				}
				break;
			default:
				;
			}
			break;
		case '\n':
			break;
		default:
			continue;
		}
	}
	fclose(in);

	return TRUE;
}

void size_reset_value(struct symbol *sym)
{
	char *newval;
	struct symbol_entry *sym_list;

	if (sym_is_const(sym))
		return;

	sym->flags |= SYMBOL_NEW | SYMBOL_CHANGED;
	sym->flags &= ~SYMBOL_VALID;
	switch (sym->type) {
	case S_INT:
	case S_HEX:
	case S_STRING:
		if (sym->user.val) {
			sym->curr.val = sym->user.val;
			sym->flags &= ~SYMBOL_NEW;
		}
		break;
	default:
		sym->user.val = NULL;
		sym->user.tri = no;
	}
}

int size_conf_choice(struct menu *menu)
{
	struct symbol *sym, *def_sym;
	struct menu *child;
	int type;
	bool is_new;

	sym = menu->sym;
	type = sym_get_type(sym);
	is_new = !sym_has_value(sym);
	if (sym_is_changable(sym)) {
		size_conf_sym(menu);
		sym_calc_value(sym);
		switch (size_get_tristate_value(sym)) {
		case no:
			return TRUE;
		case mod:
			return FALSE;
		case yes:
			break;
		}
	} else {
		switch (size_get_tristate_value(sym)) {
		case no:
			return TRUE;
		case mod:
			return FALSE;
		case yes:
			break;
		}
	}

	while (1) {
		int cnt, def;
		def_sym = sym_get_choice_value(sym);
		cnt = def = 0;
		line[0] = '0';
		line[1] = 0;
		for (child = menu->list; child; child = child->next) {
			if (!menu_is_visible(child))
				continue;
			if (!child->sym) {
				continue;
			}
			cnt++;
			if (child->sym == def_sym) {
				def = cnt;
			}
		}
		if (cnt == 1) {
			goto conf_childs;
		}
		if (!is_new) {
			cnt = def;
			sprintf(line, "%d\n", cnt);
			goto conf_childs;
		} else {
			cnt = def;
			sprintf(line, "%d\n", cnt);
			goto conf_childs;
		}
		strip(line);
		if (!line[0])
			cnt = def;
		else if (isdigit(line[0]))
			cnt = atoi(line);
		else
			continue;

conf_childs:
		for (child = menu->list; child; child = child->next) {
			if (!child->sym || !menu_is_visible(child))
				continue;
			if (!--cnt)
				break;
		}
		if (!child)
			continue;

		size_set_choice_value(sym, child->sym);
		printf("%s\n",child->sym->name);
		if (child->list) {
			indent += 2;
			size_conf_menu(child->list);
			indent -= 2;
		}
		return TRUE;
	}
}

static bool size_make_sym(struct symbol *sym, tristate val)
{
	struct menu *menu = NULL, *child = NULL;
	struct property *prop;
	struct symbol_entry *sym_list;
	struct symbol *choice_val = NULL;
	struct symbol *sym_tmp= NULL;

	if (!sym)
		return TRUE;

	/* we got choice entry's value, remember it into the list */
	if (sym_is_choice_value(sym)) {
		//printf("choice: %s\n", sym->name);
		choice_val = sym;
		sym = prop_get_symbol(sym_get_choice_prop(sym));
	}

	if (sym_is_choice(sym) && !choice_val)
		return TRUE;
	
	/* get menu entry */
	for_all_prompts(sym, prop) {
		menu = prop->menu;
		if (menu) break;
	}
	
	if (choice_val) {
		size_set_choice_value(sym, choice_val);
	} else {
		switch (sym_get_type(sym)) {
		case S_INT:
		case S_HEX:
		case S_STRING:
			break;
		case S_TRISTATE:
		case S_BOOLEAN:
			size_set_tristate_value(sym, val);
			break;
		default:
			;
		}
	}

	return TRUE;
}

static void size_reset_all_sym()
{
	int i;
	struct symbol *sym;
	for_all_symbols(i, sym) {
		size_reset_value(sym);
	}
}

static void size_output_depends(FILE *fp_index, struct symbol_entry *focus_sym)
{
	struct symbol_entry *sym_entry = NULL;
	int found = 0;
	int i;
	struct symbol *sym;
	
	for_all_config_symbols(sym_entry) {
		if(sym_entry->sym == NULL || sym_entry->sym->name == NULL)
			continue;
		if(focus_sym->type != S_MENU && sym_entry->sym == focus_sym->sym)
			continue;
		
		if(sym_entry->sym->curr.tri  == yes)
			continue;
		
		if(focus_sym->type == S_MENU) {
			if(!strcmp (focus_sym->menu_name, sym_entry->sym->name) ||\
				(sym_entry->sym->prop && sym_entry->sym->prop->text && \
				!strcmp (focus_sym->menu_name, sym_entry->sym->prop->text)))
					continue;
		}

		fprintf(fp_index, "+%s", sym_entry->sym->name);
	}
	fprintf(fp_index,"\n");
}

/* recursively make expression to no
 * if the expression is "OR", then both sub-expression should be no
 * if the expression is "AND", then only left sub-expression should be no
 */
static bool size_make_expr_no(struct expr *expr)
{
	const char *str1, *str2;
	int val1, val2;
	struct symbol_entry *sym_entry, *tmp;

	if (!expr)
		return TRUE;

	switch (expr->type) {
	case E_SYMBOL:
		size_make_sym(expr->left.sym, no);
		for_all_forward_entries(sym_entry, tmp, &config_list) {
			if(sym_entry->sym == expr->left.sym) {
				size_make_expr_no(expr->left.sym->rev_dep.expr);
				break;
			}
		}
	case E_AND:
		return size_make_expr_no(expr->left.expr);
	case E_OR:
		if (!size_make_expr_no(expr->left.expr)) {
			return FALSE;
		}
		if (!size_make_expr_no(expr->right.expr)) {
			return FALSE;
		}
		return TRUE;
	}
	return TRUE;
}

/* return the menu specified by name */
static struct menu *size_search_menu_name(struct menu *menu, char *name)
{
	struct menu *ret = NULL;
	struct property *prop = NULL;
	while (menu) {
		/* compare prompt->text */
		prop = menu->prompt;
		while(prop) {
			if (prop->text && (!strcmp(prop->text, name))) {
						return  menu;
			}
			prop = prop->next;
		}
		/* compare sym->name */
		if (menu->sym && menu->sym->name && !strcmp(menu->sym->name, name))
			return menu;

		if (menu->list) {
			ret = size_search_menu_name(menu->list, name);
			if (ret)
				break;
		}
		menu = menu->next;
	}
	return ret;
}

/* add sym under the menu to symbol_entry */
static void *size_menu_addto_symentry(struct menu *menu, struct symbol_entry **entry)
{
	struct symbol_entry *sub_list = *entry;
	struct symbol_entry *config_entry = NULL;

	while (menu) {
		if (menu->sym && menu->sym->name) {
			for_all_config_symbols(config_entry) {
				if (config_entry->sym && config_entry->sym->name) {
					if (strcmp(config_entry->sym->name, menu->sym->name) == 0) {
						size_add_sym(sub_list, config_entry->sym);
						break;
					}
				}
			}
		}
		if (menu->list) {
			size_menu_addto_symentry(menu->list, entry);
		}
		menu = menu->next;
	}
}

static struct symbol_entry *get_sub_focus_list(struct symbol_entry *focus)
{
	struct symbol_entry *menu_items;
	struct menu *menu = NULL;
	struct symbol_entry *config_entry = NULL;
	int found = 1;
	if (!focus)
		return NULL; 
	menu_items = malloc(sizeof (*menu_items));
	memset(menu_items, 0, sizeof (*menu_items));
	menu_items->next = menu_items->prev = menu_items;
	if (focus->type == S_MENU) {
		/* Get menu entry */
		menu = size_search_menu_name(&rootmenu, focus->menu_name);
		/* add menu's symbol to symbol_entry*/
		if (menu) {
			if (menu->sym && menu->sym->name) {
				for_all_config_symbols(config_entry) {
					if (config_entry->sym && config_entry->sym->name) {
						if (strcmp(config_entry->sym->name, menu->sym->name) == 0) {
							size_add_sym(menu_items, config_entry->sym);
							break;
						}
					}
				}
			}
			size_menu_addto_symentry(menu->list, &menu_items);
			if (menu_items->next == menu_items) /* not found sym */
				found = 0;
		} else {
			found = 0; /* not found menu entry */
		}
	}
	if (found == 0) {
		free(menu_items);
		return NULL;
	}
	return menu_items;
}

/*
 * Generate output files
 */
static void size_make_all_focus()
{
	FILE *fp_index, *fp_list;
	char index_path[256];
	struct symbol_entry *focus_entry = NULL;
	struct symbol_entry *config_entry = NULL;
	struct symbol_entry *output_list = NULL;
	struct symbol_entry *sub_focus_list = NULL;
	char tmp_path[MAX_PATH];
	int found;
	int found_reverse;

	focus_no = 0;
	fp_index = fopen(index_file, "w");
	if (!fp_index) {
		perror("fopen: index file");
		exit(1);
	}

	/* list file is used to hold .config file path info */
	fp_list = fopen(list_file,"w");
	if (!fp_list) {
		perror("fopen: list file");
		exit(1);
	}

	/*  copy input .config file to base .config file
	 *  no need to calc again
	 */
	sprintf(index_path, "%s" DIRDELIM "%d.config", result_path, focus_no);
	copy_asc_file(config_set, index_path);
	
	/* add title to the index file */
	fprintf(fp_index, "%s,%s,%s\n", TITLE_INDEX,TITLE_FOCUS,TITLE_ENABLED);
	fprintf(fp_index, "%d,NONE,NONE\n", focus_no);

	/* write a line into list file, absolute path is needed */
	fprintf(fp_list,"%d,%s" DIRDELIM "%d.config\n", \
					focus_no,absolute_result_path,focus_no);

	for_all_focus_symbols(focus_entry) {
		sub_focus_list = NULL;
		focus_no++;
		/* reset all symbol */
		size_reset_all_sym();

		if(focus_entry->type == S_MENU) {
			/*get sub focus list from one focus menu, for example,
			 * if focus_entry is "Device Driver", then sub_focus_list
			 * will return a symbol_entry list, each symbol_entry 
			 * contain one symbol that also enabled in .config file
			 * */
			sub_focus_list = get_sub_focus_list(focus_entry);
			if(sub_focus_list == NULL) {
				fprintf(fp_index, "%d,%s,%s\n", focus_no, focus_entry->menu_name, \
						"not enabled in .config file");
				continue;
			}
			
			for_all_config_symbols(config_entry) {
				size_make_sym(config_entry->sym,yes);
			}	
			
			struct symbol_entry *sym_entry, *tmp;
			for_all_forward_entries(sym_entry, tmp, sub_focus_list) {
				/* disable focus symbol */
				size_make_sym(sym_entry->sym, no);
		
				/* disable all reverse chain of focus symbol */
				size_make_expr_no(sym_entry->sym->rev_dep.expr);
			}
		} else {	
			/* make sure item of focus_list must aslo be in config_list */
			found = no;
		
			/* at first, enable every config item including focus symbol */
			for_all_config_symbols(config_entry) {
				size_make_sym(config_entry->sym,yes);
		 		if(config_entry->sym == focus_entry->sym){
					found = yes;
				}
			}	

			if(!found){
				fprintf(fp_index, "%d,%s,%s\n", focus_no, focus_entry->sym->name, \
						"not enabled in .config file");
				continue;
			}
		
			/* disable focus symbol */
			size_make_sym(focus_entry->sym, no);
		
			/* disable all reverse chain of focus symbol */
			size_make_expr_no(focus_entry->sym->rev_dep.expr);
		}
		
		/* calc visibility again */
		do {
			conf_cnt = 0;
			size_conf_check(&rootmenu);
		} while (conf_cnt);
	
		size_gen_config();
	
		if(focus_entry->type == S_MENU) {
			fprintf(fp_index, "%d,[%s],[%s]", focus_no, focus_entry->menu_name, \
							focus_entry->menu_name);
			// output those config items that exist in config_list, but unvisible
			size_output_depends(fp_index, focus_entry);
		} else {	
			if(focus_entry->sym->curr.tri == yes) {
				fprintf(fp_index, "%d,%s,%s\n", focus_no, focus_entry->sym->name, \
							"base_config");
			} else {
				fprintf(fp_index, "%d,%s,%s", focus_no, focus_entry->sym->name, \
						focus_entry->sym->name);
		
				// output those config items that exist in config_list, but unvisible
				size_output_depends(fp_index, focus_entry);
			}
		}
		/* write a line into list file, absolute path is needed */
		fprintf(fp_list,"%d,%s" DIRDELIM "%d.config\n",focus_no, \
						absolute_result_path,focus_no);

		if(sub_focus_list) {
				struct symbol_entry *sub_focus_entry, *temp_entry;
				for_all_forward_entries(sub_focus_entry, temp_entry, sub_focus_list) {
					size_remove_sym(sub_focus_entry);
				}
		}
	}
	size_cleanup();

	fclose(fp_index);
	fclose(fp_list);
}

static int copy_asc_file(const char *from, const char *to)
{
	char line[1024];
	char *p = NULL;
	int ret = FALSE;
	FILE *src = NULL, *dst = NULL;
	
	if (from == NULL || to == NULL)
		goto done;
	
	src = fopen(from, "r");
	if (!src)
		goto done;
	dst = fopen(to, "w");
	if (!dst)
		goto done;
	while (fgets(line, sizeof(line), src)) {
		switch(line[0]) {
			case 'C':
				if (memcmp(line, "CONFIG_", 7))
					continue;
				p = strchr(line + 7, '=');
				if (!p)
					continue;
				if (*(p + 1) == 'm')
					*(p + 1) = 'y';
				break;
			default:
				break;
		}
		//copy line to dst file
		fprintf(dst, line);
	}

	ret = TRUE;
done:
	if (src)
		fclose(src);
	if (dst)
		fclose(dst);
	
	return ret;
}

int main(int argc, char **argv)
{
	int c = 0;
	const char *kconfig_file;
	FILE *out_file = NULL, *out;
	char *arch = NULL;
	char arch_kconfig[256];

	progname = basename(argv[0]);
	size_init_lists();

	while (1) {
		int option_index = 0;
		c = getopt_long(argc, argv, "a:k:f:d:i:c:m:h", long_options, &option_index);
		if (c == -1)
			break;
		
		switch (c) {
		case 'a':
			arch = strdup(optarg);
			break;
		case 'k':
			src_tree = strdup(optarg);
			setenv(SRCTREE, (char *)src_tree, 1);
			break;
		case 'd':
			result_path = strdup(optarg);
			break;
		case 'f':
			focus_set = strdup(optarg);
			break;
		case 'i':
			index_file = strdup(optarg);
			break;
		case 'c':
			list_file = strdup(optarg);
			break;
		case 'm':
			menu_output = optarg;
			break;
		case 'h':
			size_usage();
			exit(0);
		case '?':
		default:
			size_usage();
			exit(1);
		}
	}

	if (optind < argc) {
		config_set = strdup(argv[optind]);
	} else {
		fprintf(stderr, "Missing config list file....\n");
		size_usage();
		exit(1);
	}

	/* if not set, use default value */
	if(arch == NULL)
		arch=DEFAULT_ARCH;
	if(src_tree == NULL)
		src_tree=DEFAULT_SOURCE_TREE;
	if(result_path == NULL)
		result_path=DEFAULT_OUTPUT;
	if(index_file == NULL)
		index_file=combine_path(result_path,DEFAULT_INDEX);
	if(list_file == NULL)
		list_file=combine_path(result_path,DEFAULT_LIST);

	/* get Kconfig path */
	memset(arch_kconfig,0,sizeof(arch_kconfig));
	strcpy(arch_kconfig,"arch");
	strcat(arch_kconfig, DIRDELIM);
	strcat(arch_kconfig,arch);
	strcat(arch_kconfig, DIRDELIM);
	strcat(arch_kconfig,"Kconfig");
	kconfig_file = combine_path(src_tree, arch_kconfig);

	/* if output directory doesn't exist, create it */
	if(access(result_path, F_OK) != 0){
		if(mkdir(result_path,0755)!= 0){
			perror("create output directory failure");
			exit(1);
		}	
	}

	get_current_path();
	if(result_path[0] == DIRDELIMC){
		strcpy(absolute_result_path,result_path);
	} else {
		strcat(absolute_result_path, DIRDELIM);
		strcat(absolute_result_path, result_path);
	}
	valid_stdin = isatty(0) && isatty(1) && isatty(2);
	conf_parse(kconfig_file);

	if (menu_output) {
		out = stdout;
		if (strcmp(menu_output, "-")) {
			if ((out_file = fopen(menu_output, "w")) == NULL) {
				perror("fopen");
				exit(1);
			}
			out = out_file;
		}
		size_print_menu(out);
		if (out_file)
			fclose(out_file);
	}

	/* get config_list from .config file */
	size_read_config(config_set);
	
	/* get focus_list from focus file, if focus file is not specified, 
	 * just copy config_list 
	 */
	if(focus_set){
		size_read_focus(focus_set);
	} else {
		struct symbol_entry *sym_entry, *tmp;
		for_all_forward_entries(sym_entry, tmp, &config_list) {
			size_add_sym(&focus_list, sym_entry->sym);
		}
	}
	
	size_make_all_focus();
	return 0;
}

static void size_usage()
{
	fprintf(stderr, "Usage %s <item_file> [options] \n", progname?progname:"GENCONF");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "        -a, specify arch\n");
	fprintf(stderr, "        -k, specify kernel source tree\n");
	fprintf(stderr, "        -d, specify results output directory\n");
	fprintf(stderr, "        -f, specify focus file\n");
	fprintf(stderr, "        -i, specify index file\n");
	fprintf(stderr, "        -c, specify config list file\n");
	fprintf(stderr, "        -m, specify category file\n");
}

