/*
 * KconfigSize tool---Memory tester
 * Copyright (C) 2005-2006 NEC-AS/NEC. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GPLv2, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Created by YongHua Zheng <zhengyh@sh.necas.nec.com.cn>
 * 2006-02-20 Modified by YongHua Zheng <zhengyh@sh.necas.nec.com.cn>
 */

#include "common.h"

#define DEFAULT_CONFIG_FILE "../conf/memtest.conf"

#define PROJECT "memtest"
#define AUTHORS "YongHua Zheng"

struct config_item;

typedef int (*item_set) (struct config_item *item, const char *value);

/* Structure used to parse config items in memtest.conf */
struct config_item {
#define ITEM_NULL   0
#define ITEM_STRING 1
#define ITEM_INT    2
	int type;
#define FL_NONE     0
#define FL_REQUIRE  1
	int flags;
	char *name;
	void *value;
	item_set item_set_handler;
};

/* Parse memtest.conf */
int config_parse ();

/* configure with items from memtest.conf */
int config_set (const char *name, const char *value);

/* Check validation of config items */
int config_check ();

/* Call reset helper*/
int call_resethelper ();

/* Call mem measure */
int call_memmeasure (int *child_pid);

/* Copy kernel image to specified directory */
int call_cpimage();

/* Generate tmp file name */
int tmp_get_fname ();

/* Save line number and index to tmp file */
int tmp_save (int lineno, char *index);

/* Load line number and index from tmp file */
int tmp_read (int *lineno, char **index);

/* Delete tmp file */
int tmp_remove ();

/* Create result file */
int result_create ();

/* Open result file */
int result_open ();

/* Close result file */
int result_close ();

/* Write test result to result file */
int result_write (const char *index, int type);

/* Cleanup routine */
void do_cleanup ();

/* Obtain specified line from image list file */
int get_list_buff (char **bufptr, size_t *n, int lineno);

/* Get index & image of a specified line from image list file */
int get_list_image (char **index, char **image, int lineno);

/* Set parameters used for exec reset helper or mem measure */
int item_param_set (struct config_item *item, const char *value);

/* Free parameters malloced in item_param_set */
void item_param_free (const char **param);

/* Initialized enviroment setup */
int testenv_setup (int *lineno, char **index);

/* program name */
static const char *g_prog_name;

/* path to reset helper */
static const char *g_reset_helper;

/* parameters for reset helper */
static const char **g_reset_helper_param;

/* reset acceptable keyword */
static const char *g_reset_key;

/* reset timeout value [second] */
static int g_reset_timeout;

/* reset delay time from reset acceptable */
static int g_reset_delay;

/* path to memory measure */
static const char *g_mem_measure;

/* parameters for mem measure */
static const char **g_mem_measure_param;

/* result string keywork */
static const char *g_mem_result_key;

/* path to image list file */
static const char *g_image_list;

/* path to install image */
static const char *g_install_image;

/* path to mem test result file */
static const char *g_result;

/* path to temporary file */
static const char *g_result_tmp;

/* memtest configuraion file */
static const char *g_config_file = DEFAULT_CONFIG_FILE;

static struct timeval g_last_reset_time;
static struct timeval g_reset_schedule_time;

static long g_mem_total;
static long g_mem_free;
static long g_mem_used;

static int  g_image_list_fd = -1;
static int  g_result_fd = -1;
#if !defined (BUFSIZ)
#define BUFSIZ 8192
#endif
static char g_buff[BUFSIZ];

#define C_RESET_HELPER		"RESET_HELPER"
#define C_RESET_HELPER_PARAM	"RESET_HELPER_PARAM"
#define C_RESET_KEY		"RESET_KEY"
#define C_RESET_TIMEOUT		"RESET_TIMEOUT"
#define C_RESET_DELAY		"RESET_DELAY"
#define C_MEM_MEASURE		"MEM_MEASURE"
#define C_MEM_MEASURE_PARAM	"MEM_MEASURE_PARAM"
#define C_MEM_RESULT_KEY	"MEM_RESULT_KEY"
#define C_IMAGE_LIST		"IMAGE_LIST"
#define C_INSTALL_IMAGE		"INSTALL_IMAGE"
#define C_RESULT		"RESULT"

static struct config_item g_config_items[] = {
	{ ITEM_STRING, FL_REQUIRE, C_RESET_HELPER, (void *)&g_reset_helper, NULL },
	{ ITEM_STRING, FL_NONE, C_RESET_HELPER_PARAM, (void *)&g_reset_helper_param, item_param_set },
	{ ITEM_STRING, FL_NONE, C_RESET_KEY, (void*)&g_reset_key, NULL },
	{ ITEM_INT, FL_NONE, C_RESET_TIMEOUT, (void*)&g_reset_timeout, NULL },
	{ ITEM_INT, FL_NONE, C_RESET_DELAY, (void*)&g_reset_delay, NULL },
	{ ITEM_STRING, FL_REQUIRE, C_MEM_MEASURE, (void *)&g_mem_measure, NULL },
	{ ITEM_STRING, FL_NONE, C_MEM_MEASURE_PARAM, (void *)&g_mem_measure_param, item_param_set },
	{ ITEM_STRING, FL_REQUIRE, C_MEM_RESULT_KEY, (void*)&g_mem_result_key, NULL },
	{ ITEM_STRING, FL_REQUIRE, C_IMAGE_LIST, (void *)&g_image_list, NULL },
	{ ITEM_STRING, FL_REQUIRE,C_INSTALL_IMAGE, (void *)&g_install_image, NULL },
	{ ITEM_STRING, FL_REQUIRE, C_RESULT, (void *)&g_result, NULL },
	{ ITEM_NULL, 0, NULL, NULL }
};


/**
 * left2reset_sec()
 *   return left time [sec] to reset schedule time
**/
static unsigned int left2reset_sec(void)
{
	struct timeval now;
	unsigned int left;
	
	if ( !timerisset(&g_reset_schedule_time) )
		return 0;
	
	gettimeofday(&now, NULL);
	if( timercmp(&now, &g_reset_schedule_time, >) )
		return 0;
	
	left = g_reset_schedule_time.tv_sec - now.tv_sec;
	if(g_reset_schedule_time.tv_usec > now.tv_usec)
		left++;

	return left;
}


/**
 * kill_wait_child()
 *   send SIGKILL to child and wait its termination
**/
static int kill_wait_child(pid_t child)
{
	if( child && waitpid(child, NULL, WNOHANG)==0 ){
		/* child still alive */
		kill(child, SIGKILL);
		return wait_child(child);
	}
	return 0;
}


int config_parse ()
{
	FILE *f;
	char *linebuff = NULL;
	size_t size = 0;
	ssize_t len;
	char *ptr;
	char *name, *value;

	if (!(f = fopen (g_config_file, "r")))
		return -1;

	while (!feof (f)) {
		len = getline (&linebuff, &size, f);
		if (len < 0)
			break;

		if (linebuff[0] == '#')
			continue;

		chomp (linebuff);

		if ((ptr = strchr (linebuff, '=')) == NULL)
			continue;

		*ptr++ = '\0';
		name = trim (linebuff);
		value = trim (ptr);
		if (*name == '\0')
			continue;

		if (config_set (name, value) < 0)
			return -1;
	}
	if (linebuff)
		free (linebuff);

	if (ferror (f))
		return -1;

	if (fclose (f) == EOF)
		return -1;

	return 0;
}

int config_set (const char *name, const char *value)
{
	struct config_item *item;
	char **p;

	for (item = g_config_items; item->type != ITEM_NULL; item++) {
		if (strcmp (name, item->name) == 0) {
			if (item->item_set_handler) {
				return item->item_set_handler (item, value);
			}
			else {
				if (*value == '\0')
					return 0;

				if (item->type == ITEM_STRING) {
					p = (char **)item->value;
					assert (p);
					if (*p)
						continue;
					*p = strdup (value);
					if (*p == NULL) {
						errno = ENOMEM;
						return -1;
					}
					break;
				}
				else if (item->type == ITEM_INT) {
					*(int *)item->value = atoi (value);
					break;
				}
				else assert (0);
			}
		}
	}

	return 0;
}

int config_check ()
{
	struct config_item *item;
	char **p;
	int i = 0;

	for (item = g_config_items; item->type != ITEM_NULL; item++, i++) {
		if (item->flags & FL_REQUIRE) {
			if (item->type == ITEM_STRING) {
				p = (char **)item->value;
				assert (p);
				if (*p == NULL)
					return (i + 1);
			}
		}
	}

	return 0;
}

int item_param_set (struct config_item *item, const char *value)
{
	char ***pav = (char ***)item->value;
	char *p, *c, *s;
	int ac = 2;

	assert (pav);

	if (*value == '\0')
		return 0;

	if (*pav)
		return 0;

	*pav = (char **)malloc (ac * sizeof (char *));
	if (!*pav)
		return -1;
	if (strcmp (item->name, C_RESET_HELPER_PARAM) == 0)
		(*pav)[0] = strdup (g_reset_helper ? g_reset_helper : "resethelper");
	else
		(*pav)[0] = strdup (g_mem_measure ? g_mem_measure : "memmeasure");

	if (!(*pav)[0])
		return -1;
	(*pav)[1] = NULL;

	c = (char *)value;
	s = c;
	while (1) {
		int quate = 0;
		if (*c == '"'){
			quate = 1;
			s = c+1;
		}
		else
			quate = 0;

		if (quate){
			p = strchr (s, '"');
			if (p){
				if( *(p-1)=='\\' ){
					s = p+1;
					continue;
				}
				else if( *(p+1)==' ' ){
					c++;
					*p++ = '\0';
				}
				else
					return -1;
			}
		}
		else {	/* !quate */
			p = strchr (s, ' ');
			if (p)
				*p++ = '\0';
		}
			
		ac++;
		*pav = (char **)realloc (*pav, ac * sizeof (char *));
		if (!*pav)
			return -1;
		if (!((*pav)[ac - 2] = strdup (c)))
			return -1;
		if (p) {
			c = trim (p);
			s = c;
		}
		else break;
	}

	if (ac > 2)
		(*pav)[ac -1] = NULL;

	return 0;
}

void item_param_free (const char **param)
{
	int i;

	if (param) {
		for (i = 0;param[i]; i++)
		 	free ((void *)param[i]);

		free (param);
	}
}

int result_create ()
{
	int fd;
	size_t len;

	fd = open (g_result, O_CREAT | O_WRONLY | O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd < 0)
		return -1;

	len = sprintf (g_buff, "%s, %s, %s, %s\n", "index", "used",
			"free", "total");
	if (full_write (fd, g_buff, len) != len)
		return -1;

	close (fd);

	return 0;
}

int result_open ()
{
	if (g_result_fd == -1) {
		g_result_fd = open (g_result, O_WRONLY | O_APPEND);
		if (g_result_fd < 0)
			return -1;
		if (fcntl (g_result_fd, F_SETFD, FD_CLOEXEC) < 0)
			return -1;
	}
	return 0;
}

int result_write (const char *index, int type)
{
	size_t len;

	switch (type) {
		case 0:
			sprintf (g_buff, "%s, %d, %d, %d\n", index,
				g_mem_used, g_mem_free, g_mem_total);
			break;
		case 1:
			sprintf (g_buff, "%s, %s\n", index, "MEM_FAIL");
			break;
		case 2:
			sprintf (g_buff, "%s, %s\n", index, "CP_FAIL");
			break;
		default:
			assert (0);
	}

	len = strlen (g_buff);
	if (full_write (g_result_fd, g_buff, len) != len)
		return -1;

	return 0;
}

int result_close ()
{
	if (g_result_fd != -1) {
		close (g_result_fd);
		g_result_fd = -1;
	}

	return 0;
}

int get_list_buff (char **bufptr, size_t *n, int lineno)
{
	static int line = 1;
	ssize_t len;
	
	if (g_image_list_fd == -1) {
		g_image_list_fd = open (g_image_list, O_RDONLY);
		if (g_image_list_fd < 0)
			return -1;
		if (fcntl (g_image_list_fd, F_SETFD, FD_CLOEXEC) < 0)
			return -1;

		line = 1;
	}

	while (1) {
		len = readline (g_image_list_fd, (void **)bufptr, n);
		if (len > 0) {
			if (line++ < lineno)
				continue;
			break;
		}
		else
			break;
	}

	return len;
}

int get_list_image (char **index, char **image, int lineno)
{
	static char *read_buf = NULL;
	static size_t n = 0;
	int len, strno;
	char *ptr, *buff;

	len = get_list_buff (&read_buf, &n, lineno);
	if (len > 0) {
		chomp (read_buf);
		buff = read_buf;
		strno = 0;
		while ((ptr = strsplit (&buff, ",")) != NULL && strno < 2) {
			switch (strno) {
				case 0:
					if (*index)
						free (*index);
					*index = strdup (ptr);
					if (*index == NULL)
						return -1;
					break;
				case 1:
					if (*image)
						free (*image);
					*image = strdup (ptr);
					if (*image == NULL)
						return -1;
					break;
				default:
					assert (0);
			}
			strno++;
		}
	}
	else if (len == 0) {
		if (read_buf)
			free (read_buf);
		n = 0;
		return 0;
	}
	else return -1;

	if (strno < 2)
		return -2;

	return len;
}

int tmp_get_fname ()
{
	size_t len = 0;

	assert (g_result);

	if (g_result_tmp)
		return 0;

	stradd (&g_result_tmp, g_result, &len);
	stradd (&g_result_tmp, ".tmp", &len);

	return 0;
}

/* Load current line number & index from tmp file */
int tmp_read(int *lineno, char **index)
{
	FILE *f;
	char *line_buf = NULL;
	size_t n = 0;
	char *buff, *ptr;
	int strno;

	if (!(f = fopen (g_result_tmp, "r"))) {
		return -1;
	}

	if (!feof (f)) {
		if (getline (&line_buf, &n, f) < 0)
			return -1;

		chomp (line_buf);
		buff = line_buf;
		strno = 0;
		while ((ptr = strsplit (&buff, ",")) && strno < 2) {
			switch (strno) {
				case 0:
					*lineno = atoi (ptr);
					break;
				case 1:
					if (*index)
						free (*index);
					*index = strdup (ptr);
					if (!*index)
						return -1;
					break;
				default:
					assert (0);
			}
			strno++;
		}
	}

	if (line_buf)
		free (line_buf);

	if (ferror (f))
		return -1;

	if (fclose (f) == EOF)
		return -1;

	assert (strno > 1);

	return 0;
}

/* Current line number & index will be saved in tmp file */
int tmp_save (int lineno, char *index)
{
	FILE *f;

	assert (index);

	if (!(f = fopen (g_result_tmp, "w")))
		return -1;

	fprintf (f, "%d, %s\n", lineno, index);
	fclose (f);

	sync ();
	sleep (1);

	return 0;
}

/* Remove the temp file */
int tmp_remove ()
{
	unlink (g_result_tmp);
	errno = 0; /* Discards the error */
	return 0;
}

int call_resethelper ()
{
	int ret = 0;
	
	if (!g_reset_helper) 
		return -1;
		
	if (!g_reset_helper_param)
		ret = execl_command (g_reset_helper, g_reset_helper, NULL);
	else
		ret = execv_command (g_reset_helper, g_reset_helper_param);
	
	if (ret)
		return ret;
	
	return 0;
	
}

/**
 * get_key_string()
 *   Try to get string starting with expected keyword
 * params
 *   fd	(I)	: file discriptor to be read
 *   key (I)	: expected keyword
 *   pbuf (I/O)	: pointer to read buffer,
 *   		  string including the keyword will be contained here
 *   pn (I/O)	: buffer size
 * returns
 *   < 0	: error
 *   GKS_TIMEOUT: timeout occured
 *   GKS_GET	: got the expected string
**/
#define	GKS_TIMEOUT	0
#define	GKS_GET		1

static int get_key_string(int fd, const char *key, char **pbuf, size_t *pn)
{
	fd_set fds;
	struct timeval tv;
	struct timeval *ptv;
	int res;
	int len;
	size_t key_len;

	key_len = strlen(key);
	
	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	
	while (1) {
		if ( timerisset(&g_reset_schedule_time) ){
			timerclear(&tv);
			tv.tv_sec = left2reset_sec();
			ptv = &tv;
		}
		else
			ptv = NULL;

		res = select(fd + 1, &fds, NULL, NULL, ptv);
		if (res < 0) {
			if (errno != EINTR) {
				return -1;
			}
			sleep(1);
			continue;
		} 
		if (res == 0)
			return GKS_TIMEOUT;

		len = readline (fd, (void **)pbuf, pn);
		if (len <= 0)
			return -1;
		fprintf (stdout, "%s", *pbuf);

		if ( strncmp(key, *pbuf, key_len) == 0) {
			/* the string has come */
			return GKS_GET;
		}
	}
}

/*
 * the function returns value as following:
 * 1. >0  success
 * 2. =0  timeout
 * 3. <0  failure
 *
 * child_pid returns value as following: 
 * 1. <0  child process exited  [ no need to kill it any more ]
 * 2. >0  the pid of child process [ child process still exists ] 
 */
int call_memmeasure (int *child_pid)
{
	pid_t pid;
	int p[2];
	int len;
	char *read_buf = NULL;
	size_t n = 0;
	char *ptr, *buff;
	int strno;
	int res;
	int ret = 0;

	if (pipe (p) < 0)
		return -1;

	switch (pid = fork ()) {
		case 0:
			/* 
			 * child process, redirect the stdout
			 * to pipe
			 */
			close (p[0]);
			close (1);
			dup2 (p[1], 1);
			if (!g_mem_measure_param) {
				execl (g_mem_measure, g_mem_measure, NULL);
				perror ("execl");
			}
			else {
				execv (g_mem_measure, g_mem_measure_param);
				perror ("execv");
			}
			exit (1);
		case -1:
			return -1;
		default:
			break;
	}
	/* child process will never reach here */
	close (p[1]);
	
	/* try to get memory measurement result */
	ret = get_key_string(p[0], g_mem_result_key, &read_buf, &n);
	kill(SIGTERM, pid);
	if( ret <= 0){	/* error, set reset schedule soon */
		timerclear(&g_reset_schedule_time);
		goto end;
	}
	
	chomp (read_buf);
	buff = read_buf + strlen(g_mem_result_key);
	strno = 0;
	while ((ptr = strsplit (&buff, ",")) && (strno < 3)) {
		switch (strno) {
			case 0:
				g_mem_total = atoi (ptr);
				break;
			case 1:
				g_mem_used = atoi (ptr);
				break;
			case 2:
				g_mem_free = atoi (ptr);
				break;
			default:
				assert (0);
		}
		strno++;
	}
	if (strno < 3) {
		ret = -1;
		goto end;
	}
	
	/* try to get reset accept and set reset schedule */
	res = 1;
	if(g_reset_key)
		res = get_key_string(p[0], g_reset_key, &read_buf, &n);
	
	if ( g_reset_delay && (res > 0) ){
		/* in case of ...
		   - reset delay is specified
		     and
		     reset accept keyword has come
		   OR
		   - reset delay is specified
		     and
		     reset accept keyword not specified
		*/
		gettimeofday(&g_reset_schedule_time, NULL);
		g_reset_schedule_time.tv_sec += g_reset_delay;
	}
	else{
		/* in case of ...
		   - reset accept keyword wait timed out
		   OR
		   - reset delay not specified
		*/
		timerclear(&g_reset_schedule_time);
	}
	
	end :
	close (p[0]);
	free (read_buf);

	/*judge whether child process still exists*/
	if(waitpid(pid, NULL, WNOHANG)==0 )
		*child_pid = pid;
	else
		*child_pid = -1;

	return (int)ret;
}

int call_cpimage (const char *srcfile) {
	int srcfd;
	int dstfd;
	size_t len;
	int ret = 0;

	if ((srcfd = open (srcfile, O_RDONLY)) < 0) {
		return -1;
	}

	if ((dstfd = open (g_install_image, O_CREAT | O_WRONLY | O_TRUNC,
				S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {
		close (srcfd);
		return -1;
	}

	while (1) {
		len = full_read (srcfd, g_buff, sizeof (g_buff));
		if (len == (size_t) -1) {
			ret = -1;
			break;
		}
		else if (len == 0)
			break;

		if (full_write (dstfd, g_buff, len) != len) {
			ret = -1;
			break;
		}
	}

	close (srcfd);
	close (dstfd);

	sync ();
	sleep (2);

	return ret;
}

void do_cleanup ()
{
#define DO_FREE(mem)				\
	do {					\
		if (mem)			\
			free ((void *)mem);	\
	} while (0)

	DO_FREE(g_reset_helper);
	DO_FREE(g_reset_key);
	DO_FREE(g_mem_measure);
	DO_FREE(g_mem_result_key);
	DO_FREE(g_image_list);
	DO_FREE(g_install_image);
	DO_FREE(g_result);
	DO_FREE(g_result_tmp);
#undef DO_FREE
	item_param_free (g_reset_helper_param);
	item_param_free (g_mem_measure_param);
	if (g_image_list_fd != -1)
		close (g_image_list_fd);
	if (g_result_fd != -1)
		close (g_result_fd);
}

/* 0 means normal
 * -1 means result file can't be accessed
 * -2 means all image files can't be copied, so no need to proceed.
 */
int testenv_setup (int *lineno, char **index)
{
	char *image = NULL;
	int ret;
	int success = 0;

	*lineno = 1;

	if (result_create () < 0)
		return -1;

	while (1) {
		ret = get_list_image (index, &image, *lineno);
		if (ret < 0) {
			if (ret == -2) {
				warn (NULL, "File %s, Lineno %d: Invalid format\n",
						g_config_file, *lineno);
				*lineno++;
				continue;
			}
			return -1;
		}
		else if (ret == 0)
			break;

		if (call_cpimage (image) < 0) {
			if (result_open () < 0)
				return -1;

			if (result_write (*index, 2) < 0)
				return -1;

			result_close ();
			*lineno++;
			continue;
		}

		success ++; 
		if (tmp_save (*lineno, *index) < 0)
			return -1;

		if (call_resethelper () < 0) {
			return -1;
		}

		/* record new reset time and next timeout */
		gettimeofday(&g_last_reset_time, NULL);
		if (g_reset_timeout){
			g_reset_schedule_time = g_last_reset_time;
			g_reset_schedule_time.tv_sec += g_reset_timeout;
		}else
			timerclear(&g_reset_schedule_time);
		break;
	}

	if(success > 0)
		return 0; 
	else 
		return -2;
}

void usage (int status) {
	if (status != 0)
		fprintf (stderr, "Try %s --help for mor information\n",
				g_prog_name);
	else {
		printf ("Usage: %s <OPTION>\n", g_prog_name);
		fputs ("\
Kernel memory measurement.\n\
\n\
        -c, --config      specifying configuration file\n\
        -f, --fail        start up in fail mode\n\
        -h, --help        display help\n\
        -v, --version     display version\n", stdout);
	}

	exit (status);
}

static struct option memtest_options[] = {
	{"help", 0, 0, 'h'},
	{"config", 1, 0, 'c'},
	{"fail", 0, 0, 'f'},
	{"version", 0, 0, 'v'},
	{0, 0, 0, 0}
};

enum {
	MODE_NORMAL = 0,
	MODE_FAIL   = 1
};

int main (int ac, char *av[])
{
	int option_index;
	int c;
	int mode = MODE_NORMAL;
	int lineno;
	char *cur_index = NULL;
	char *index = NULL;
	char *image = NULL;
	int mem_pid = 0;
	int strno, ret;

	if ((g_prog_name = strrchr (av[0], '/')) != NULL)
		g_prog_name++;
	else
		g_prog_name = av[0];

	while (1) {
		c = getopt_long (ac, av, "fc:vh", memtest_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
			case 'c':
				g_config_file = strdup (optarg);
				if (!g_config_file) {
					perror ("strdup");
					ret = 1;
					goto end;
				}
				break;

			case 'f':
				mode = MODE_FAIL;
				break;

			case_GETOPT_VERSION_CHAR (PROJECT, AUTHORS)

			case_GETOPT_HELP_CHAR

			default:
				usage (1);
				break;
		}
	}

	if (config_parse () < 0) {
		perror ("config_parse");
		ret = 1;
		goto end;
	}

	c = config_check ();
	if (c > 0) {
		fprintf (stderr, "Error in %s: %s was not set\n",
				g_config_file, g_config_items[++c].name);
		ret = 1;
		goto end;
	}

	if (tmp_get_fname () < 0) {
		perror ("tmp_get_fname");
		ret = 1;
		goto end;
	}

	if (mode == MODE_NORMAL || tmp_read (&lineno, &cur_index) < 0) {
		if (mode == MODE_FAIL && errno != ENOENT) {
			perror ("tmp_read");
			ret = 1;
			goto end;
		}
		errno = 0;

		if (testenv_setup (&lineno, &cur_index) < 0) {
			perror ("testenv_setup");
			ret = 1;
			goto end;
		}
	}

	if (result_open () < 0) {
		perror ("result_open");
		ret = 1;
		goto end;
	}
	while (1) {
		if (mode == MODE_FAIL)
			mode = MODE_NORMAL;
		ret = call_memmeasure (&mem_pid);

		if (ret <= 0) {
			warn (NULL, "memory measurement failed\n");
			result_write (cur_index? cur_index : index, 1);
		}
		else
			result_write (cur_index? cur_index : index, 0);

		if (cur_index) {
			free (cur_index);
			cur_index = NULL;
		}
get_next:
		while (1) {
			lineno++;
			ret = get_list_image (&index, &image, lineno);
			if (ret < 0) {
				if (ret == -2) {
					warn (NULL, "File %s, Lineno %d: Invalid format\n",
							g_config_file, lineno);
					continue;
				}
				perror ("get_list_image");
				ret = 1;
				goto end;
			} else break;
		}
		if (ret == 0)
			break;

		if (call_cpimage (image) < 0) {
			if (result_write (index, 2) < 0) {
				perror ("result_write");
				ret = 2;
				goto end;
			}
			goto get_next;
		}

		if (tmp_save (lineno, index) < 0) {
			perror ("tmp_save");
			ret = 1;
			goto end;
		}

		/* if reset_delay is specified, wait for the rest time.
		 * the rest time is calculated in left2reset_sec(). 
		 */
		sleep( left2reset_sec() );

		/* if child process exists, just kill it*/
		if(mem_pid > 0)
			kill_wait_child(mem_pid);
		
		mem_pid = 0;
		
		if (call_resethelper () < 0) {
			fprintf (stderr, "Excution of reset helper failed\n");
			ret = 1;
			goto end;
		}

		/* record new reset time and next timeout */
		gettimeofday(&g_last_reset_time, NULL);
		if (g_reset_timeout){
			g_reset_schedule_time = g_last_reset_time;
			g_reset_schedule_time.tv_sec += g_reset_timeout;
		}else
			timerclear(&g_reset_schedule_time);
	}
	
	tmp_remove ();
	
end :
	if(mem_pid > 0)
		kill_wait_child(mem_pid);
	
	do_cleanup();

	return ret;
}
